# import wandb
import random
import os
import gc

# from ..utils import WebdavManager
from collections import defaultdict
from distiller.quantization import DorefaQuantizer
from gym.spaces import Box, Discrete
from pathlib import Path
from qpd.config import Config
from qpd.networks.wrapper.stable_baselines.sb_wrapper import SBWrapper
from qpd.utils import summary
from qpd.utils.evaluator import Evaluator
from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.vec_env.base_vec_env import VecEnv
from torch.distributions import Categorical
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Dict, Optional, Union

from .memory import Memory

from ..networks import *
from ..networks.wrapper.model_wrapper import ModelWrapper
# from ..utils import test_network, merge_config

class PolicyDistillation:
    def __init__(self, teacher: ModelWrapper, student: ModelWrapper, config: Config): #, webdav_manager: Optional[WebdavManager] = None):
        self.teacher = teacher
        self.student = student
        self.config = config

        self.epochs = self.config.compression_config.epochs
        self.batch_size = self.config.compression_config.batch_size
        self.device = self.config.compression_config.device
        self.check_point_frequency = self.config.compression_config.checkpoint_frequency
        self.mem_update_frequency = self.config.memory_config.update_frequency
        
        # self.webdav_manager = webdav_manager
        self.results = defaultdict(list)
        
        self.student.to(self.device)
        self.optimizer = torch.optim.RMSprop(self.student.parameters(), lr=config.student_config.learning_rate)
        self.evaluator = Evaluator(config, verbose=True)
        self.quantizer = self.get_quantizer()
        
        self.initial_environment = self.config.environment_constructor()
        self.best_student = None
        
        self.memory_order = list(range(1, self.config.compression_config.epochs+1))
        self.memory = self.load_memory()
        random.shuffle(self.memory_order)


    def train(self):
        best =  -1 * float("inf")
        for epoch in range(self.epochs):
            print(f"Start epoch {epoch}")
            epoch_loss = 0

            self.student.train()
            self.student.to(self.device)
            
            with torch.enable_grad():
                for states, outputs in self.memory.sample(self.batch_size):
                    loss = self.continuous_loss(states.to(self.device, dtype=torch.float), outputs.to(self.device)) if self.config.env_continuous_action_space else self.discrete_loss(states.to(self.device, dtype=torch.float), outputs.to(self.device))
                    
                    assert not torch.isnan(loss), "Error: loss can't be nan!"
                    epoch_loss += self.step(loss)

            # wandb.log({"loss": epoch_loss, "Epoch": epoch})
            print(f"Epoch {epoch} completed, loss: {epoch_loss}")
            self.results["Loss"].append({"Epoch": epoch, "Value": epoch_loss})
            if (epoch and epoch % self.config.evaluator_config.student_test_frequency == 0) or self.config.evaluator_config.student_test_frequency == 1:
                best, is_best = self.test(epoch, best)
                if is_best:
                    self.save_student("best")

            if (epoch and epoch % self.check_point_frequency == 0) or self.check_point_frequency == 1:
                self.save_student(epoch)

            if epoch and epoch % self.mem_update_frequency == 0:
                self.load_memory(epoch, self.memory)

        return self.best_student
        # if self.webdav_manager:
        #     self.webdav_manager.upload_results(self.results)

    def load_memory(self, epoch: int=0, memory: Optional[Memory]=None) -> Memory:
        memory_index = (self.memory_order.pop() if epoch else 0)
        # memory_index = (epoch if epoch else self.memory_order.pop())
        path = f"{self.config.data_directory}/Resources/{str(self.initial_environment.envs[0].env) if self.config.is_vec_env else str(self.initial_environment.env)}/Memory/{self.teacher.__class__.__name__}/sd_{str(self.config.evaluator_config.student_driven)}/{memory_index}"
        if os.path.exists(path):
            tmp_memory = Memory(self.config, meta_class=self.teacher, path=path)

            if memory:
                memory.load_memory(tmp_memory)
                return memory
            return tmp_memory
        else:
            print()
            print("= Memory ==========================================================")
            print(f"= Could not find memory with index: {memory_index}, generating new data...")
            if not memory:
                memory = Memory(self.config, meta_class=self.teacher)
                steps = memory.max_size
            else:
                steps = self.config.memory_config.update_size

            print(f"= Collecting data for {steps}")
            results = self.evaluator.run(self.student, steps=steps, collect=True, teacher=self.teacher) if self.config.evaluator_config.student_driven else self.evaluator.run(self.teacher, steps=steps, collect=True)
            print("= Current average steps:", results["Steps"])
            print("= Current average reward:", results["Reward"])
            print("= Average amount of episode:", results["EpisodeCount"])
            print("===================================================================")
            print()

            if "GPULab" in os.environ or self.config.evaluator_config.student_driven:
                for _ in range(len(results["Memory"])):
                    memory.load_memory(results["Memory"].pop())
                    gc.collect()

            else:
                total_size = sum([len(m) for m in results["Memory"]])
                tmp_memory = Memory(self.config, size=total_size, meta_class=self.teacher)
                for _ in range(len(results["Memory"])):
                    tmp_memory.load_memory(results["Memory"].pop())
                    gc.collect()
                tmp_memory.save(path)
                if epoch:
                    memory.load_memory(tmp_memory)
                else:
                    return tmp_memory
            return memory


    def discrete_loss(self, states:torch.Tensor, teacher_outputs:torch.Tensor) -> torch.Tensor:
        T = self.config.compression_config.T
        student_policy, student_value, _ = self.student(states)
        
        teacher_policy, teacher_value_std = teacher_outputs[:, :student_policy.shape[1]], teacher_outputs[:, student_policy.shape[1]:]
        teacher_value = teacher_value_std[:,:teacher_policy.shape[1]]

        if self.config.compression_config.categorical:
            dist = Categorical(logits=student_policy)
            student_policy = dist.probs
            dist2 = Categorical(logits=teacher_policy)
            teacher_policy = dist2.probs

        policy_loss = nn.KLDivLoss(reduction="batchmean")(F.log_softmax(student_policy, dim=1), F.softmax(teacher_policy/T, dim=1)) #type: ignore
        critic_importance = self.config.compression_config.critic_importance
        if critic_importance and not student_value == None:
            value_loss = nn.HuberLoss(reduction="mean")(student_value.to(self.device), teacher_value.to(self.device))
            value_sum = value_loss.sum().detach()
            policy_sum = policy_loss.sum().detach()
            if policy_sum < 1e-6 or value_sum < 1e-6: # Avoid divide by 0
                return (1-critic_importance)  * policy_loss + critic_importance * value_loss
            else:
                return ((1-critic_importance) * (policy_loss / policy_sum) + critic_importance * (value_loss / value_sum)) * (value_sum + policy_sum)
        else:
            return policy_loss
        

    def continuous_loss(self, states:torch.Tensor, teacher_outputs:torch.Tensor) -> torch.Tensor:
        self.student.to(self.device)
        student_policy, _, student_std = self.student(states)
        # student_policy, student_value = student_outputs[:, :actions], student_outputs[:, actions:]
        teacher_policy, teacher_value_std = teacher_outputs[:, :student_policy.shape[1]], teacher_outputs[:, student_policy.shape[1]:]


        if self.config.compression_config.distribution == "Std":
            teacher_std = teacher_value_std[:, -teacher_policy.shape[1]:] # type: ignore

            if not teacher_std.gt(0).all():
                teacher_std[teacher_std==0] = 1
                print("Fixing teacher std")
            if not student_std.gt(0).all():
                student_std[student_std==0] = 1
                print("Fixing student std")

            if self.config.compression_config.loss == "KL":
                policy_loss = torch.log((teacher_std / student_std)) + (student_std ** 2 + (student_policy - teacher_policy) ** 2) / (2 * (teacher_std ** 2)) - 1/2 # type: ignore
            elif self.config.compression_config.loss == "Huber":
                policy_loss = nn.HuberLoss(reduction="none")(student_policy, teacher_policy)
                std_loss = nn.HuberLoss(reduction="none")(student_std, teacher_std)
                policy_loss += std_loss
            elif self.config.compression_config.loss == "MSE":
                policy_loss = nn.MSELoss(reduction="none")(student_policy, teacher_policy)
                std_loss = nn.MSELoss(reduction="none")(student_std, teacher_std)
                policy_loss += std_loss
            else:
                raise RuntimeError("Unknown loss type: " + self.config.compression_config.loss)
        else:
            if self.config.compression_config.loss == "Huber":
                policy_loss = nn.HuberLoss(reduction="none")(student_policy, teacher_policy) #type: ignore
            elif self.config.compression_config.loss == "MSE":
                policy_loss = nn.MSELoss(reduction="none")(student_policy, teacher_policy) #type: ignore
            else:
                raise RuntimeError("Unsupported combination of distribution and loss!")

        return policy_loss.sum()


    def step(self, loss: torch.Tensor) -> float:
        loss.backward()
        self.optimizer.step()
        self.optimizer.zero_grad()
        self.student.zero_grad()

        if self.quantizer:
            self.quantizer.quantize_params()

        return loss.item()


    def get_quantizer(self) -> Optional[DorefaQuantizer]:
        if self.config.quantization_config.enabled:
            bits = self.config.quantization_config.bits
            # self.student = distiller.modules.convert_model_to_distiller_lstm(self.student)
            quantizer = DorefaQuantizer(self.student, self.optimizer, bits_activations=bits, bits_weights=bits, bits_bias=bits)
            quantizer.prepare_model()
            quantizer.quantize_params()
            return quantizer
        return None


    def save_student(self, epoch: Union[int, str], name: str="student"):
        checkpoint_directory = self.config.data_directory
        base_path = f"{checkpoint_directory}/Data/{self.config.run_name}/compression/{self.config.compression_config.starting_time}"
        Path(f"{base_path}/{epoch}").mkdir(parents=True, exist_ok=True)

        self.student.save(f"{base_path}/{epoch}/{name}")

        # if self.webdav_manager:
        #     self.webdav_manager.upload_model(student_save_path, epoch, not self.config["testing"]["local_checkpoints"])
        
    def test(self, epoch: int, best: Optional[float]=None) -> Optional[float]:
        is_best = False
        print()
        print("= Testing =========================================================")
        results = self.evaluator.run(self.student, episodes=self.config.evaluator_config.episodes)
        self.student.to(self.config.compression_config.device)
        log = {"Epoch": epoch}
        for metric in ["Steps", "Reward"]:
            self.results[metric].append({"Epoch": epoch, "Value": results[metric]})
            log[metric] = results[metric]

        # wandb.log(log)

        print("= Current average steps:", results["Steps"])
        print("= Current average reward:", results["Reward"])
        print("= Average amount of episode:", results["EpisodeCount"])
        print("===================================================================")
        print()

        if best and results["Reward"] > best:
            # wandb.log({"Best Reward": results["Reward"]})
            best = results["Reward"]
            is_best = True

        return best, is_best
    
    def get_action_space_shape(self):
        action_space = self.student.policy.action_space
        is_box = isinstance(action_space, Box)
        assert is_box or isinstance(action_space, Discrete), "[ERROR]: Unsupported action space!"

        if is_box:
            return action_space.shape
        else:
            return action_space.n

        
