import datetime

from gymnasium.spaces import Box, Discrete

import qpd.networks.wrapper.model_wrapper as wrapper
import qpd.utils.model_utils as model_utils
import torch
import torch.nn as nn

from qpd.networks.wrapper.student.fully_connected_student import FCStudentNet
from stable_baselines3.common.vec_env.base_vec_env import VecEnv

class AbstractConfig:
    key = None

    def init(self, config: dict):
        if self.key in config.keys():
            self.update(config[self.key])
        return self

    def update(self, config):
        self.__dict__.update((k, config[k]) for k in set(config).intersection(self.__dict__))

    def serializable(self) -> dict:
        tmp = {}

        for key, value in self.__dict__.items():
            v = None

            if isinstance(value, AbstractConfig):
                v = value.serializable()
            elif isinstance(value, type):
                v = str(value)
            elif callable(value):
                continue
            else:
                v = value

            tmp[key] = v

        return tmp

class MemoryConfig(AbstractConfig):
    key = "memory"
    def __init__(self):
        self.size = 54000
        self.update_frequency = 1
        self.update_size = 5400
        self.device = "cpu"
        self.check_consistency = True
        self.frame_stack_optimization = False

class StudentNetConfig(AbstractConfig):
    key = "student"
    def __init__(self):
        self.student_network = FCStudentNet
        self.learning_rate = 1e-4
        self.activation_func = nn.ReLU
        self.extra_student_kwargs = {}

class CompressionConfig(AbstractConfig):
    key = "compression"
    def __init__(self):        
        self.checkpoint_frequency = 2
        self.starting_time = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        self.epochs = 600
        self.batch_size = 256
        self.T = 0.01
        self.device = torch.device("cuda")
        self.critic_importance = 0.5
        self.categorical = False
        self.loss = "MSE"
        self.distribution = "Std"

class EvaluatorConfig(AbstractConfig):
    key = "evaluator"
    def __init__(self):
        self.environment_is_done_filter = lambda env_info: True
        self.student_test_frequency = 1
        self.episodes = 50
        self.initialize = 30
        self.env_subproc = False
        self.env_workers = 1
        self.ray_workers = 24
        self.device = "cpu"
        self.deterministic = False
        self.student_driven = False
        self.exploration_rate = 0.

class QuantizationConfig(AbstractConfig):
    key = "quantization"
    def __init__(self):
        self.enabled = False
        self.bits = 8

class Config(AbstractConfig):
    def __init__(self, environment_constructor, config={}):
        self.environment_constructor = lambda: environment_constructor(self)
        
        self.memory_config = MemoryConfig().init(config)
        self.student_config = StudentNetConfig().init(config)
        self.compression_config = CompressionConfig().init(config)
        self.evaluator_config = EvaluatorConfig().init(config)
        self.quantization_config = QuantizationConfig().init(config)

        self.verbose = True
        self.data_directory = "."
        self.run_name = "Test"

        # AUTO-DEFINED
        self.model_has_critic = False
        self.observation_shape = None
        self.action_shape = None
        self.output_shape = None
        self.is_vec_env = False
        self.env_continuous_action_space = False

        self.update(config)

    def update_memory_config(self, **kwargs):
        self.memory_config.update(**kwargs)

    def update_student_config(self, **kwargs):
        self.student_config.update(**kwargs)

    def update_compressor_config(self, **kwargs):
        self.compression_config.update(**kwargs)
    
    def update_evaluator_config(self, **kwargs):
        self.evaluator_config.update(**kwargs)

    def update_quantization_config(self, **kwargs):
        self.quantization_config.update(**kwargs)

    def init_env_model_based_params(self, wrapper: "wrapper.ModelWrapper"):
        wrapper.to(self.compression_config.device)
        initialized_environment = self.environment_constructor()
        
        action_space = initialized_environment.action_space
        is_box = isinstance(action_space, Box)
        assert is_box or isinstance(action_space, Discrete), f"[ERROR]: Unsupported action space of type {type(action_space)}!"
        
        state = initialized_environment.reset()     # Get state to determin shape of observation
        state = wrapper.observation_to_tensor(state).to(self.compression_config.device).float()   # get usable state and determin if env is vec_env

        self.is_vec_env = isinstance(initialized_environment, VecEnv)
        
        if self.is_vec_env:
            self.observation_shape = state.shape[1:]
        else:
            self.observation_shape = state.shape

        if is_box:
            self.env_continuous_action_space = True
        else:
            self.env_continuous_action_space = False

        action, value, action_std = wrapper(state)

        self.model_has_critic = True if value is not None else False

        outputs = model_utils.get_output(action[0], value[0] if value is not None else None, action_std[0] if action_std is not None else None) if self.is_vec_env else model_utils.get_output(action, value, action_std)

        self.action_shape = action.shape[1:] if self.is_vec_env else action.shape
        self.output_shape = outputs.shape