import torch

from stable_baselines3.common.base_class import BaseAlgorithm
from stable_baselines3.common.distributions import (
    BernoulliDistribution,
    CategoricalDistribution,
    DiagGaussianDistribution,
    MultiCategoricalDistribution,
    StateDependentNoiseDistribution
)
from stable_baselines3.common.policies import BasePolicy
import qpd.networks.wrapper.stable_baselines.sb_wrapper as sbw

class WrapperONPActorCritic(sbw.SBWrapper):
    def forward(self, obs):
        features = self.policy.extract_features(obs)
        if self.policy.share_features_extractor:
            self.latent_pi, latent_vf = self.policy.mlp_extractor(features)
        else:
            pi_features, vf_features = features
            self.latent_pi = self.policy.mlp_extractor.forward_actor(pi_features)
            latent_vf = self.policy.mlp_extractor.forward_critic(vf_features)
        
        self.mean_actions = self.policy.action_net(self.latent_pi)
        value = self.policy.value_net(latent_vf)
        
        action_std = None

        if self.config.compression_config.distribution == "Std":
            if isinstance(self.policy.action_dist, StateDependentNoiseDistribution):
                variance = torch.mm(self.latent_pi ** 2, self.policy.action_dist.get_std(self.policy.log_std) ** 2)
                action_std = torch.sqrt(variance + self.policy.action_dist.epsilon)

            elif isinstance(self.policy.action_dist, DiagGaussianDistribution):
                action_std = torch.ones_like(self.mean_actions) * self.policy.log_std.exp()
            elif isinstance(self.policy.action_dist, CategoricalDistribution):
                action_std = None   # In case of discrete action space
            else:
                raise RuntimeError("Invalid distribution type!")

        return self.mean_actions, value, action_std
    
    def _get_actions(self):
        distribution = None

        if isinstance(self.policy.action_dist, DiagGaussianDistribution):
            distribution = self.policy.action_dist.proba_distribution(self.mean_actions, self.policy.log_std)
        elif isinstance(self.policy.action_dist, CategoricalDistribution):
            # Here mean_actions are the logits before the softmax
            distribution = self.policy.action_dist.proba_distribution(action_logits=self.mean_actions)
        elif isinstance(self.policy.action_dist, MultiCategoricalDistribution):
            # Here mean_actions are the flattened logits
            distribution = self.policy.action_dist.proba_distribution(action_logits=self.mean_actions)
        elif isinstance(self.policy.action_dist, BernoulliDistribution):
            # Here mean_actions are the logits (before rounding to get the binary actions)
            distribution = self.policy.action_dist.proba_distribution(action_logits=self.mean_actions)
        elif isinstance(self.policy.action_dist, StateDependentNoiseDistribution):
            distribution = self.policy.action_dist.proba_distribution(self.mean_actions, self.policy.log_std, self.latent_pi)
        else:
            raise ValueError("Invalid action distribution")

        return distribution.get_actions(self.config.evaluator_config.deterministic)     
