from abc import abstractmethod
import json
import torch
import time
import numpy
import random
import ray

from typing import Dict, List, Optional
from torch.distributions import Normal

from qpd.config import Config
from qpd.networks.wrapper.model_wrapper import ModelWrapper
from qpd.utils.model_utils import get_output
from ..compression.memory import Memory
from dataclasses import dataclass, field


@dataclass
class Worker:
    idx: int = 0
    action_queue: List[int] = field(default_factory=list)
    memory: Optional[Memory] = None
    finished: bool = False
    steps: int = 0
    rewards: int = 0
    repeats: int = 0

    def reset(self, number: int):
        self.action_queue = [0 for _ in range(int(numpy.round(random.random() * number)))]
        self.steps = 0
        self.rewards = 0
        self.repeats = 0


@dataclass
class Stats:
    repeats: int = 0
    frames: int = 0
    effective_frames: int = 0
    steps: List[int] = field(default_factory=list)
    rewards: List[int] = field(default_factory=list)

    def update(self, worker: Worker):
        self.repeats += worker.repeats
        self.effective_frames += worker.steps
        self.steps.append(worker.steps - worker.repeats)
        self.rewards.append(worker.rewards)


class Evaluator:
    def __init__(self, config: Config, verbose: bool=False) -> None:
        self.config = config
        # self.env = create_env(config)
        self.verbose = verbose

        ray.init(
            _system_config={
                "local_fs_capacity_threshold": 0.9999999999,
                "object_spilling_config": json.dumps(
                    {
                        "type": "filesystem", 
                        "params": {
                            "directory_path": f"{config.data_directory}/spill"
                        }
                    },
                )
            },
        )

    def run(self, network: torch.nn.Module, collect: bool=False, steps: int=0, episodes: int=0, teacher=None) -> Dict:
        network.eval()
        network.to(torch.device(self.config.evaluator_config.device))

        @ray.remote(num_gpus=0.1)
        def run_ray(function, *args, **kwargs):
            return function(*args, **kwargs)

        test_function = Evaluator.test_vec if self.config.is_vec_env else Evaluator.test

        if self.config.evaluator_config.ray_workers:
            worker_steps = list(map(lambda x: len(x), numpy.array_split(numpy.arange(steps), self.config.evaluator_config.ray_workers)))
            worker_episodes = list(map(lambda x: len(x), numpy.array_split(numpy.arange(episodes), self.config.evaluator_config.ray_workers)))
            old_metrics = ray.get(
                [run_ray.remote(
                    test_function, 
                    self.config, 
                    network,
                    steps=s, 
                    episodes=e, 
                    collect=collect, 
                    verbose=self.verbose,
                    teacher=teacher
                    ) for s, e in zip(worker_steps, worker_episodes)]
                )
            metrics = {item: sum(list(map(lambda x: x[item], old_metrics)), []) for item in ["Memory"]}
            metrics.update({item: numpy.mean(list(map(lambda x: x[item], old_metrics))) for item in ["Reward", "Steps", "EpisodeCount"]})
            # metrics = {item: sum(list(map(lambda x: x[item], old_metrics)), []) for item in ["Memory", "Reward", "Steps"]}
        else:
            metrics = test_function(
                self.config, 
                network, 
                steps=steps, 
                episodes=episodes, 
                collect=collect,
                verbose=self.verbose,
                teacher=teacher)
            
        network.train()
        return metrics
    
    @abstractmethod
    def test(config: Config, 
            model_wrapper: ModelWrapper,
            steps: int=0, 
            episodes: int=0, 
            collect: bool=False, 
            verbose=True,
            teacher: ModelWrapper=None) -> Dict:
        
        env = config.environment_constructor()
        # model = None
        # model = model_wrapper.original_type(type(model_wrapper.policy), env, device=config.evaluator_config.device, **config.build_student_kwargs())
        # model.policy = model_wrapper.policy

        if teacher:
            teacher.to(device=config.evaluator_config.device)

        total_rewards = []
        total_steps = []
        frames = 0
        effective_frames = 0

        with torch.no_grad():
            env_state = env.reset()

            worker_action_queue = [0 for _ in range(int(numpy.round(random.random() * config.evaluator_config.initialize)))]
            worker_steps_since_last_reward = 0
            worker_steps = 0
            worker_finished = False

            worker_memory = None
            if collect:
                worker_memory = Memory(config, size=int(steps) + 10000)

            start = time.time()
            if verbose:
                print(f"Start workers")
            while not numpy.all(worker_finished):
                current_actions = []

                state, _ = model_wrapper.observation_to_tensor(env_state)
                action, value, action_std = model_wrapper(state)

                if len(worker_action_queue) > 0:
                    current_actions.append(worker_action_queue.pop())
                else:
                    current_actions = model_wrapper.get_actions()
                    
                if collect:                    
                    if teacher and teacher is not model_wrapper:
                        action, value, action_std = teacher(state)

                    if not worker_finished:
                        if worker_memory.will_overflow:
                            print(f"Growing memory to size {worker_memory.max_size + 5000}!")
                            worker_memory.grow(5000)

                        worker_memory.update(
                            state.unsqueeze(0),
                            get_output(action, value, action_std), 
                            worker_steps == 0
                        )

                env_state, reward, done, infos = env.step(current_actions) #type:ignore

                frames += 1

                worker_steps += 1
                if reward > 0:
                    worker_steps_since_last_reward = 0
                else:
                    worker_steps_since_last_reward += 1

                if done:
                    if config.evaluator_config.environment_is_done_filter(infos):
                        assert "episode" in infos, "No episode key in infos dict from gym environment"
                        total_rewards.append(infos["episode"]["r"])
                        worker_action_queue = [0 for _ in range(int(numpy.round(random.random() * config.evaluator_config.initialize)))]
                        effective_frames += worker_steps
                        total_steps.append(worker_steps)
                        worker_steps = 0

                        if (steps and effective_frames >= steps) or (episodes and len(total_rewards)):
                            worker_finished = True

        if collect:
            worker_memory.prune()

        if verbose:
            end = time.time()
            print("Average and effective framerate:", round(frames / (end-start), 3), round(effective_frames / (end-start), 3))

        return {"Steps": numpy.mean(total_steps) if len(total_steps) else 0, "Reward": numpy.mean(total_rewards) if len(total_rewards) else 0, "Memory": worker_memory if collect else []}

    @abstractmethod
    def test_vec(config: Config, 
            model_wrapper: ModelWrapper,
            steps: int=0, 
            episodes: int=0, 
            collect: bool=False, 
            verbose=True,
            teacher: ModelWrapper=None) -> Dict:
        
        env = config.environment_constructor()
        
        model_wrapper.to(device=config.evaluator_config.device)
        if teacher:
            teacher.eval()
            teacher.to(device=config.evaluator_config.device)

        total_rewards = []
        total_steps = []
        frames = 0
        effective_frames = 0

        workers = config.evaluator_config.env_workers

        with torch.no_grad():
            state = env.reset()
            worker_action_queue = [[0 for _ in range(int(numpy.round(random.random() * config.evaluator_config.initialize)))] for _ in range(workers)]
            workers_steps_since_last_reward = [0 for _ in range(workers)]
            worker_steps = [0 for _ in range(workers)]
            worker_finished = [False for _ in range(workers)]

            worker_memories = []
            if collect:
                worker_memories = [Memory(config, size=int(steps / workers) + 10000) for _ in range(workers)]

            start = time.time()
            if verbose:
                print(f"Start workers")
            while not numpy.all(worker_finished):
                current_actions = []
                state = model_wrapper.observation_to_tensor(state).float()
                action, value, action_std = model_wrapper(state)

                if max([len(q) for q in worker_action_queue]) > 0:
                    for worker in range(workers):
                        if len(worker_action_queue[worker]):
                            current_actions.append(worker_action_queue[worker].pop())
                else:
                    current_actions = model_wrapper.get_actions()
                    
                if collect:
                    if teacher and teacher is not model_wrapper:
                        action, value, action_std = teacher(state)

                    for worker in range(workers):
                        if not worker_finished[worker]:
                            if worker_memories[worker].will_overflow():
                                print(f"Growing memory to size {worker_memories[worker].max_size + 5000}!")
                                worker_memories[worker].grow(5000)

                            outputs = get_output(action[worker], value[worker] if value is not None else None, action_std[worker] if action_std is not None else None)

                            worker_memories[worker].update(state[worker].unsqueeze(0), outputs, worker_steps[worker] == 0)

                state, reward, done, infos = env.step(current_actions) #type:ignore

                frames += workers
                for worker in range(workers):
                    worker_steps[worker] += 1
                    if reward[worker] > 0:
                        workers_steps_since_last_reward[worker] = 0
                    else:
                        workers_steps_since_last_reward[worker] += 1

                    if done[worker] and not worker_finished[worker]:
                        # print(f"Done: {infos}")
                        if config.evaluator_config.environment_is_done_filter(infos[worker]):
                            assert "episode" in infos[worker]
                            total_rewards.append(infos[worker]["episode"]["r"])
                            worker_action_queue[worker] = [0 for _ in range(int(numpy.round(random.random() * config.evaluator_config.initialize)))]
                            effective_frames += worker_steps[worker]
                            total_steps.append(worker_steps[worker])
                            worker_steps[worker] = 0

                            if (steps and effective_frames >= steps) or (episodes and len(total_rewards) + workers - 1 >= episodes):
                                if verbose:
                                    print(f"Finished with {effective_frames} steps and {len(total_rewards)} episodes!")
                                worker_finished[worker] = True


        if collect:
            for worker in range(workers):
                worker_memories[worker].prune()

        if verbose:
            end = time.time()
            print("Average and effective framerate:", round(frames / (end-start), 3), round(effective_frames / (end-start), 3))

        return {"Steps": numpy.mean(total_steps) if len(total_steps) else 0, "Reward": numpy.mean(total_rewards) if len(total_rewards) else 0, "Memory": worker_memories if collect else [], "EpisodeCount": len(total_rewards)/workers}
